/*
 * enclib.h
 *
 * This file is part of the ttf2pk package.
 *
 * Code <-> Postscript tables
 *
 *    Loyer Frederic <loyer@ensta.fr>
 */

enum _EncodingScheme
{
  encUnicode,
  encMac,
  encFontSpecific
};
typedef enum _EncodingScheme EncodingScheme;

void set_encoding_scheme(EncodingScheme e);

char *code_to_adobename(int code);
int adobename_to_code(char *s);


/* end */
