LIBRARY LIBTIFF
EXETYPE WINDOWS
CODE PRELOAD MOVEABLE DISCARDABLE
DATA PRELOAD MOVEABLE MULTIPLE
HEAPSIZE 4096
STACKSIZE 8192
EXPORTS
_TIFFGetVersion             @1
_TIFFClose                  @2
_TIFFFlush                  @3
_TIFFFlushData              @4
_TIFFGetField               @5
_TIFFVGetField              @6
_TIFFGetFieldDefaulted      @7
_TIFFVGetFieldDefaulted     @8
_TIFFReadDirectory          @9
_TIFFScanlineSize           @10
_TIFFStripSize              @11
_TIFFVStripSize             @12
_TIFFTileRowSize            @13
_TIFFTileSize               @14
_TIFFVTileSize              @15
_TIFFFileno                 @16
_TIFFGetMode                @17
_TIFFIsTiled                @18
_TIFFIsByteSwapped          @19
_TIFFCurrentRow             @20
_TIFFCurrentDirectory       @21
_TIFFCurrentStrip           @22
_TIFFCurrentTile            @23
_TIFFReadBufferSetup        @24
_TIFFLastDirectory          @25
_TIFFSetDirectory           @26
_TIFFSetSubDirectory        @27
_TIFFUnlinkDirectory        @28
_TIFFSetField               @29
_TIFFVSetField              @30
_TIFFWriteDirectory         @31
_TIFFReadScanline           @32
_TIFFWriteScanline          @33
_TIFFReadRGBAImage          @34
_TIFFOpen                   @35
_TIFFFdOpen                 @36
_TIFFClientOpen             @37
_TIFFFileName         		 @38
_TIFFError                  @39
_TIFFWarning                @40
_TIFFSetErrorHandler        @41
_TIFFSetWarningHandler      @42
_TIFFComputeTile            @43
_TIFFCheckTile              @44
_TIFFNumberOfTiles          @45
_TIFFReadTile               @46
_TIFFWriteTile              @47
_TIFFComputeStrip           @48
_TIFFNumberOfStrips         @49
_TIFFReadEncodedStrip       @50
_TIFFReadRawStrip           @51
_TIFFReadEncodedTile        @52
_TIFFReadRawTile            @53
_TIFFWriteEncodedStrip      @54
_TIFFWriteRawStrip          @55
_TIFFWriteEncodedTile       @56
_TIFFWriteRawTile           @57
_TIFFSetWriteOffset         @58
_TIFFSwabShort              @59
_TIFFSwabLong               @60
_TIFFSwabArrayOfShort       @61
_TIFFSwabArrayOfLong        @62
_TIFFReverseBits            @63
_TIFFGetBitRevTable         @64
_TIFFModeCCITTFax3          @65
