/*-----------------------------------------------------------------
LOG
    GEM - Graphics Environment for Multimedia

    A sphere

    Copyright (c) 1997-1998 Mark Danks. mdanks@cybermed.ucsd.edu
    For information on usage and redistribution, and for a DISCLAIMER OF ALL
    WARRANTIES, see the file, "GEM.LICENSE.TERMS" in this distribution.

-----------------------------------------------------------------*/

#ifndef INCLUDE_SPHERE_H_
#define INCLUDE_SPHERE_H_

#include "Base/GemGluObj.h"

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
    sphere
    
    Creates a sphere

KEYWORDS
    geo

DESCRIPTION
        
-----------------------------------------------------------------*/
class sphere : public GemGluObj
{
    CPPEXTERN_HEADER(sphere, GemGluObj)

    public:

        //////////
        // Constructor
    	sphere(t_floatarg size);
    	
    protected:
    	
    	//////////
    	// Destructor
    	virtual ~sphere();

    	//////////
    	// Do the rendering
    	virtual void 	render(GemState *state);
};

#endif	// for header file
