@echo off

echo Installing GEM...

del ..\examples\gem*.pd
copy examples\* ..\examples > tempInstall.trash
del ..\doc\gem*.*
copy doc\gem*.* ..\doc > tempInstall.trash
copy help\* ..\test > tempInstall.trash

del tempInstall.trash

echo Install finished.
