{***************************************************************************}
{*                                                                         *}
{* The Portable TrueType Engine - Copyright 1995, 1996 David TURNER        *}
{*_________________________________________________________________________*}
{*                                                                         *}
{*                                                                         *}
{*   TTError : A very simple unit to define the engine's error messages    *}
{*             a the global 'Error' variable                               *}
{*                                                                         *}
{*                                                                         *}
{***************************************************************************}

unit TTError;

interface

uses FreeType;

  procedure Check_Error( error : Integer );

  procedure Panic1( message : String );
  procedure Trace1( message : String );

(* The Pascal version of the library doesn't support multiple       *)
(* threads. We use a global error variable, called simply "error"   *)
(* to report all defects. The various functions return an error     *)
(* condition, which can be either Success (false) or Failure (true) *)

(* Note that the use of macros in the C version to automate error   *)
(* reporting makes the two source trees very similar, even if they  *)
(* differ from some design points like this one                     *)

var
  error : integer;

implementation


  procedure Panic1;
  begin
    writeln( message );
    halt(1);
  end;


  procedure Trace1;
  begin
    writeln( message );
  end;


  procedure Check_Error;
  var
    num : String[4];
  begin
    if error <> TT_Err_Ok then
    begin
      str( -error:3, num );
      Panic1( 'Error code = ' + num );
    end;
  end;

end.

